# Motivated by the Effective TCL "Scrollable Form" demo.

package require Efftcl
package require Ttk

# The variable keyword() is a list of entry definitions for particular
# values. Each element is itself a 6-element list defining the PBS
# option (switch) associated with it (if any), a fancy name for the
# keyword, text to be used with tool tip help, the type of entry,
# defaults for setting up the entry (depends on type), and a tag
# indicating what category the keyword belongs in. The indices are
# defined by:

set idx(keyswitch) 0
set idx(keyname) 1
set idx(keytext) 2
set idx(keytype) 3
set idx(keydefault) 4
set idx(keytag) 5

# Keyword categories are defined by 3 values:

set idx(required) 0
set idx(optional) 1
set idx(advanced) 2

# Available entry control types are defined by 5 values:

set idx(entry_field) 0
set idx(spin_box) 1
set idx(combo_box) 2
set idx(dir_select) 3
set idx(file_select) 4

# The hosts data array is organized as a list of lists. Each entry is
# a 3-element list. The elements define the host name, available
# queues and their characteristics, and the default queue
# settings. The indices therefore are:

set idx(hostname) 0
set idx(hostqueues) 1
set idx(hostdefaults) 2

# Each queue is organized as a list of lists, with each queue entry
# being a 3-element list. The elements define the queue name, the
# queue limits, and any other resources available to the queue. The
# indices are thus:

set idx(queuename) 0
set idx(queuelimits) 1
set idx(queueresources) 2

# The queue limits are 3-element lists containing the maximum wall
# time, the maximum number of nodes per job, and the allowed PPN
# value.  The indices are thus:

set idx(qmaxwalltime) 0
set idx(qmaxnodes) 1
set idx(qppn) 2

# The host default list also is a 3-element list. The elements define
# the system default queue, a default node count, and a default run
# time. Thus the indices are:

set idx(defaultq) 0
set idx(defaultnodes) 1
set idx(defaulttime) 2

# Each element is a 3-element list of host name, the list of queues,
# and the defaults:

set hosts [list \
   [list mike \
	[list [list workq [list 72 128 16] [] ] \
	     [list checkpt [list 72 200 16] [] ] \
	     [list single [list 72 1 1] [] ] \
	     [list gpu [list 24 16 16] [] ] \
	     [list bigmem [list 48 2 16] [] ] \
	     [list bigmemtb [list 48 1 40] [] ] \
	     [list lasigma [list 72 28 16] [] ] \
	     [list mwfa [list 72 8 16] [] ] ] \
	[list workq 1 1] ] \
   [list philip \
	[list [list workq [list 72 4 8] [] ] \
	     [list single [list 168 1 1] [] ] \
	     [list checkpt [list 72 4 8] [] ] \
	     [list gpu [list 24 2 8] [] ] \
	     [list bigmem [list 48 2 8] [list mem48 mem96] ] ] \
	[list workq 1 1] ] \
]

# The script generator should know about 3 different types of jobs:

set jobtypes [list serial mpi openmp]
					   
# Customize available keywords. Start with a list of all keyword
# names, then define the settings for each keyword. The order here
# determines the order of appearance.

set keywords [list user_name job_name allocation job_type procs_per_task \
		  nodes walltime queue workdir task start infiles pbsfile \
		  email_options email_address datetime chkpt_interval \
		  join_list keep_list stdout_path stderr_path priority \
		  restartable shell_path_list env_list full_env \
		  suppress_jobid resource]

# Specify a bold font:

set bold "Helvetica 10 bold"

set pbsTemplate(head) "#! /bin/bash
#######################################################################
# Begin WQ prologue section. This file was autogenerated by wq.qui.
#######################################################################
#PBS -A %s
#PBS -l nodes=%d:ppn=%d%s
#PBS -l walltime=%s
#PBS -q %s
#PBS -N %s
"

set pbsTemplate(prologue) "
# Set number of workers per node:

WPN=%d

# Set the working directory:

WORKDIR=%s

# Name of the file containing the list of input files:

FILES=\${WORKDIR}/%s

# Name of the PBS script file:

JOBFILE=\${WORKDIR}/%s

# Set the starting line in the file. Allows you to skip over pervious
# completed tasks. The default is 1 (i.e. start from the beginning).

START=%d

# Name of the task script each worker is expected to run to process
# the files sent to it as the only command line argument.

TASK=\${WORKDIR}/%s
"

set pbsTemplate(epilogue) "
#
# Begin WQ epilogue section.
# What follows is the main WQ script.  It should be considered powerful
# magic. Dabbled with at your own peril.
########################################################################

# Drop into the working directory.

cd \${WORKDIR}

# Only the mother superior has PBS_JOBID defined, so we will be
# passing it to the other nodes as \$2. Use this fact to decide if
# we are running on the mother superior or a compute node:

if \[ \"\${2}x\" = \"x\" ] ; then

   # Must be running on the mother superior. Remember our host name.

   MS=`uname -n`

   # Use a bit of magic to strip off the trailing host name and
   # leave only the job number from PBS_JOBID:

   JOBNUM=\${PBS_JOBID%%.*}
   HOSTLIST=\${WORKDIR}/hostlist.\${JOBNUM}

   # We want the mother superior host name first. So, take the host
   # list provided, sort it into a unique list of names, with MS first.
   # This assures it's node ID, or position in the hostlist, is 1.

   echo \${MS} > \${HOSTLIST}
   grep -v \${MS} \${PBS_NODEFILE} | uniq | sort >> \${HOSTLIST}

   # Make a local copy of the PBS script since only the mother superior
   # can see it at job start.

   JOBFILE=\${WORKDIR}/pbs.\${JOBNUM}
   cp \$0 \$JOBFILE
   chmod a+x \${JOBFILE}

   # Mother superior must start up the dispatcher, so:

   python \${WORKDIR}/wq.py --start \$START --dispatcher \${TASK} \
       --inputs \${FILES} &

   # Give it a chance to spin up since the dispatcher must be ready
   # to accept connections from the workers upon request.

   sleep 5

   # Ready to start the script on all compute nodes. This will fire up
   # workers. We'll pass PBS_WALLTIME and the job number as arguments.
   # They'll connect to the dispatcher and start work immediately.

   for H in `cat \${HOSTLIST}` ; do
      if \[ \${H} != \${MS} ] ; then
         ssh -n \${H} \${JOBFILE} \${PBS_WALLTIME} \${JOBNUM} &
      fi
   done

   # Finally, mother superior can also start workers:

   python \${WORKDIR}/wq.py --workers \${WPN} --mothersuperior \${MS} \
       --time \${PBS_WALLTIME}

   # Make sure to wait until all the processes are done!

   wait

else

   # Must be running on a compute node. The job number was passed by the
   # mother superior (see above).

   HOSTLIST=\${WORKDIR}/hostlist.\$2

   # Now, we have to get the name of mother superior from the host
   # list. Thats so we know where the dispatcher is running. Simply
   # grab the first entry from the hostlist file and press on.

   MS=`head -1 \${HOSTLIST}`

   # Ready to go. Spin up the workers. The mother superior passed
   # the job wall time as argument 1 when the script is called, so
   # we have all the values needed for workers:

   python \${WORKDIR}/wq.py --workers \${WPN} --mothersuperior \${MS} \
       --time \$1

fi

# Give a bit of time to make sure dispatcher has shut down cleanly.
# The WAIT above should allow for this, but coming down on the side of
# paranoia:

sleep 2
"

# The serial template contains no format conversions.

set taskTemplate(serial) "#! /bin/bash
#
# This script was auto-generated for a serial WQ task.

# First, split out the parts of the input file name.

FILE=\$1
DIR=`dirname \${FILE}`
BASE=`basename \${FILE}`

# Customize the enviroment with any additional path or variable
# settings here. Set CMD to the command line string to be executed.
#
# CMD=\"echo \\\"This is a test\\\"\"

cd \$DIR

# Use \"if false\" for testing, or \"if true\" for production.

if true ; then
   eval \"\${CMD}\"
else
   echo \"DIR=\${DIR}; BASE=\${BASE}; CMD=\${CMD}\"
   echo \"Faking It On Hosts: \${HOSTLIST}\"
   sleep 2
fi
"

# The mpi template contains one format conversion for the number of
# processes per MPI task.

set taskTemplate(mpi) "#! /bin/bash
# This script was auto-generated for an MPI WQ task.
#
# Split up the parts of the input file name:

FILE=\$1
DIR=`dirname \${FILE}`
BASE=`basename \${FILE}`

# Set number of processors per task, and prep the host list.

PROCS=%s
HOSTNAME=`uname -n`
HOSTLIST=\"\"
for i in `seq 1 \${PROCS}`; do
   HOSTLIST=\"\${HOSTNAME},\${HOSTLIST}\"
done 
HOSTLIST=\${HOSTLIST%%,*}

# Customize here for whatever is needed to run the desired command, such
# as adding paths and setting environment variables.
#
# CMD should contain the command exactly as one would write it on the
# command line. For example:
#
# CMD=\"mpirun -host \${HOSTLIST} -np \${PROCS} app < \${FILE} > \${BASE}.out\"

CMD=

cd \$DIR

# Use \"if false\" for testing, or \"if true\" for production.

if true ; then
   eval \"\${CMD}\"
else
   echo \"\${CMD}\"
   echo \"Faking It On Hosts: \${HOSTLIST}\"
   sleep 2
fi
"

# This template contains one format conversion for the number of threads
# per process.

set taskTemplate(openmp) "#! /bin/bash
#
# This script was auto-generated for an OpenMP WQ task.
#
# Set the number of threads per process.

export OMP_NUM_THREADS=%d

# Split out the various parts of the input file name.

FILE=\$1
DIR=`dirname \${FILE}`
BASE=`basename \${FILE}`

# Customize here for any additions to paths or other environment
# settings. Then build the command line. CMD needs to be set to a string
# that appears exactly as one would type on the command line. A complex
# example might look like:
#
# CMD=\"\${BLASTN} -task blastn -outfmt 7 -max_target_seqs 1\"
# CMD=\"\${CMD} -num_threads \${OMP_NUM_THREADS}\"
# CMD=\"\${CMD} -db /project/giant/foo/fum\"
# CMD=\"\${CMD} -query \${FILE}\"
# CMD=\"\${CMD} -out \${DIR}/IMG_genome_blast.\${BASE}\"

CMD=

# Use \"if false\" for testing, \"if true\" for production.

if false ; then
   eval \"\${CMD}\"
else
   echo \"\${CMD}\"
   # This just slows things way down for testing.
   sleep 1
fi
"


proc Debug { msg } {

    # Handy function to generate debugging message boxes. Any values
    # to display can be incorporated when the msg text is provided.

    tk_messageBox -title "DEBUG" -icon info -type ok -message "$msg"

}


proc wqSettings { } {

    # Set up all the constants based on the hostname specified in "host".
    # The variable keyword is loaded with lists of 6 elements each that
    # define the widget settings and other data to associate with the
    # keyword. The function of the elements varies some, depending on the
    # type of widget desired, but basically the elements are:
    #
    #   1. The PBS option string, if any.
    #   2. The fancy name for the keyword.
    #   3. A help string to load into the widget tool tip.
    #   4. The type of widget to use for the keyword.
    #   5. Widget specific settings, such as list of values and defaults.
    #   6. Category setting - required, optional, or advanced.

    global keyword
    global keydefaultvals
    global host
    global hosts
    global qlims
    global qresources
    global idx
    global jobtypes

    set settings ""
    set found 0

    # Locate and extract the varies lists for this host.

    foreach h $hosts {

	if { [string compare [lindex $h $idx(hostname)] $host] == 0 } {

	    set qlist [lindex $h $idx(hostqueues)]
	    set queues []
	    set dqueue [lindex [lindex $h $idx(hostdefaults)] $idx(defaultq)]
	    set dnodes [lindex [lindex $h $idx(hostdefaults)] $idx(defaultnodes)]
	    set dtime [lindex [lindex $h $idx(hostdefaults)] $idx(defaulttime)]

	    foreach q $qlist {

		set name [lindex $q $idx(queuename)]
		lappend queues $name
		set qlims($name) [lindex $q $idx(queuelimits)]
		set qresources($name) [lindex $q $idx(queueresources)]

	    }

	    set found 1
	    break

	}
    }

    if { $found == 0 } {

	tk_messageBox -title "UNKNOWN HOST" -icon error -type ok \
	    -message "Host \"$host\" not defined. Exiting"
	exit 1

    }

    # Collect widgets for all the required keywords here.  

    set keyword(job_type) \
	[list "" "Job Type" "Type of processing required" \
	     $idx(combo_box) [list $jobtypes serial] $idx(required)] 

    # Consider setting default from the USER env variable.

    set keyword(user_name) \
	[list "-u" "User Name" "Name used to log onto system" \
	     $idx(entry_field) "" $idx(required)]

    # Way to check on default allocation? POLICE file? "balance"?

    set keyword(allocation) \
	[list "-A" "Allocation" "Charge code for job" \
	     $idx(entry_field) "" $idx(required)]

    # Number of nodes to request. Will be checked against queue constraints.

    set n [lindex $qlims($dqueue) $idx(qmaxnodes)]
    set keyword(nodes) \
	[list "-l nodes=" "Nodes" "Number of nodes" \
	     $idx(spin_box) [list [list 1 $n] 1] $idx(required)]

    # Wall time set up.

    set s [Time2Secs $dtime]
    set s [Secs2Time $s]
    set keydefaultvals(walltime) $s
    set keyword(walltime) \
	[list "-l walltime=" "Walltime" \
	     "Maximum wall time to allow job to run. (hh, hh:mm, hh:mm:ss)" \
	     $idx(entry_field) $s $idx(required)]
    
    # Process/threads per task.
    
    set n [lindex $qlims($dqueue) $idx(qppn)]

    if { [string compare $dqueue single] == 0 } {

	set n 1

    }

    set keyword(procs_per_task) \
	[list "" "Procs Per Task" \
	     "Desired number of process/threads per task." \
	     $idx(combo_box) [list [list 1 $n] $n] $idx(required)]

    # No reasonable default for job name except what PBS provides

    set keyword(job_name) \
	[list "-N" "Job Name" "Name to assign to job" \
	     $idx(entry_field) WQJob $idx(required)]

    # Queue list is host specific.

    set keyword(queue) \
	[list "-q" "Destination Queue" "Name of queue to run job in" \
	     $idx(combo_box) [list $queues $dqueue] $idx(required)]

    set keyword(workdir) \
	[list "" "Work Directory" "Name of directory to run job in" \
	     $idx(dir_select) "" $idx(required)]

    set keyword(task) \
	[list "" "Task Command" "Command to control a single task" \
	     $idx(file_select) wq.sh $idx(required)]

    set keyword(infiles) \
	[list "" "Input File List" "List of input file names to process" \
	     $idx(file_select) wq.lst $idx(required)]

    set keyword(pbsfile) \
	[list "" "PBS Script Name" "Name of PBS job script" \
	     $idx(file_select) wq.pbs $idx(required)]

    # Collect all the optional keywords here:

    set keyword(start) \
	[list "" "Start" "Starting record" \
	     $idx(entry_field) "1" $idx(optional)]

    set keyword(resource) \
	[list "-l" "Queue Resources" "Special resource requests" \
	     $idx(combo_box) [list $qresources($dqueue) \
			     [lindex $qresources($dqueue) 0]] $idx(optional)]

    set keyword(email_options) \
	[list "-m" "Email Options" \
	     [concat "n (none); or any combination of: " \
		  "b (job begin); e (job end); a (job abort)"] \
	     $idx(combo_box) [list [list n b e a be ba ea bea] n] \
	     $idx(optional)]

    # Any way to determine user default email when logged in?

    set keyword(email_address) \
	[list "-M" "Email Address" \
	     "Destination email address for notifications" \
	     $idx(entry_field) "" $idx(optional)]

    set keyword(join_list) \
	[list "-j" "Join Output" \
	     "n (none); o (stdout); e (stderr); oe (both)" \
	     $idx(combo_box) [list [list n o e oe] n] $idx(optional)]

    set keyword(keep_list) \
	[list "-k" "Keep files on node" \
	     "n (none); o (stdout); e (stderr); oe (both)" \
	     $idx(combo_box) [list [list n o e oe] n] $idx(optional)]

    set keyword(stdout_path) \
	[list "-o" "Stdout File Path" \
	     [concat "File path name to hold standard output.  " \
		  "Default: jobname.oNNNNNN"] \
	     $idx(entry_field) "" $idx(optional)] 

    set keyword(stderr_path) \
	[list "-e" "Stderr File Path" \
	     [concat "File path name to hold standard error.  " \
		  "Default: jobname.eNNNNNN"] \
	     $idx(entry_field) "" $idx(optional)]

    set keyword(env_list) \
	[list "-v" "Environment Variable List" \
	     "List of environment variables (var\[=val]\[,var\[=val],...])" \
	     $idx(entry_field) "" $idx(optional)]

    set keyword(full_env) \
	[list "-V" "Pass Full Environment" "Pass all variables to job" \
	     $idx(combo_box) [list [list n y] n] $idx(optional)]

    # Collect all the advanced keywords here. They are used only in
    # special circumstances.

    set keyword(suppress_jobid) \
	[list "-z" "Suppress JobID Output" \
	     "Do not emit job ID on submission" \
	     $idx(combo_box) [list [list n y] n] $idx(advanced)]

    set keyword(datetime) \
	[list "-d" "Allowed Start Time" \
	     [concat "Time job is eligible to start - " \
		  "\[\[cc]yy]mmddhhmm\[.ss] Default: immediately)"] \
	     $idx(entry_field) "" $idx(advanced)]

    # Need fancy way to allow the c=mm version.

    set keyword(chkpt_interval) \
	[list "-c" "Checkpoint Interval" \
	     [concat "Time between checkpoints.  n (none); " \
		  "s (at shutdown); c (queue - default); " \
		  "c=mm (every mm minute)"] \
	     $idx(entry_field) "" $idx(advanced)]

    set keyword(priority) \
	[list "-p" "Job priority" "-1024 to 1024" \
	     $idx(spin_box) [list [list -1024 1024] 0] $idx(advanced)]
    
    set keyword(restartable) \
	[list "-r" "Restartable" "Job can be restarted" \
	     $idx(combo_box) [list [list n y] n] $idx(advanced)]

    set keyword(shell_path_list) \
	[list "-S" "Shell Path List" \
	     "Path to desired shell \n(path\[@host]\[,,path\[@host],, ... ])" \
	     $idx(entry_field) "" $idx(advanced)]

}


proc Time2Secs { t } {

    # Convert hh, hh:mm or hh:mm:ss to seconds.  Throw away any
    # fractions so we can deal with integer arithmetic.

    set v [lindex [split [split $t "."] :] 0]
    set s 0

    switch [llength $v] {
	1 { set s [expr [lindex $v 0] * 3600]}
	2 { set s [expr [lindex $v 0] * 3600 + [lindex $v 1] * 60] }
	3 { set s [expr [lindex $v 0] * 3600 + [lindex $v 1] * 60 \
		       + [lindex $v 2]]}
	default {  tk_messageBox -icon error -type ok \
		       -message "\"$t\" not a valid time!" }
    }
}


proc Secs2Time { s } {

    # Convert seconds to hh:mm:ss. Throw away any fractions so we can
    # deal with integer arithmetic. 

    set v [lindex [split $s .] 0]
    set h [expr $v / 3600]
    set v [expr $v % 3600]

    if { $v > 0 } {

	set m [expr $v / 60]
	set s [expr $v % 60]

    } else {

	set m 0
	set s 0

    }

    return [format "%02d:%02d:%02d" $h $m $s]
}


proc wqCheck { key } {

    # Attempt to validate keyword values, taking dependencies into
    # account. Mostly it just returns 1, but a few cases return 0 if
    # there is no automatic way to handle and if messages would be
    # annoying. This allows for checking only when absolutely needed.

    global keyvals
    global keydefaultvals
    global keywidgets
    global qlims
    global queuelimits
    global qresources
    global idx

    set rtn 0

    switch $key {

	procs_per_task {

	    # Depends on job type, so run a job_type check.

	    wqCheck job_type
	    set rtn 1
	}

	job_type {

	    # May check is for procs_per_task of 1 if job type is serial.

	    if { [string compare $keyvals(job_type) serial] == 0 } {

		set keyvals(procs_per_task) 1

	    }
	}

	walltime { 

	    # Time can't be negative, or greater than the max time allowed
	    # in a queue.

	    set maxtime [Time2Secs [lindex $qlims($keyvals(queue)) \
					$idx(qmaxwalltime)]]
	    set have [Time2Secs $keyvals(walltime)]

	    if { $have < 0 } {

		tk_messageBox -title "NEGATIVE TIME" -icon info -type ok \
		    -message [concat "Resetting negative value to " \
				  "$keydefaultvals(walltime)"]
		set have [Time2Secs $keydefaultvals(walltime)]

	    } else {

		if { $have > $maxtime } { set have $maxtime }

	    }

	    set keyvals(walltime) [Secs2Time $have]
	    set rtn 1

	}

	queue {

	    # Value is constrained by list selection, so just have to
	    # make sure walltime and maximum nodes are set correctly.

	    wqCheck walltime

	    # Reset maximum node count selectable.

	    set w [lindex $keywidgets(nodes) 0]
	    $w configure -to [lindex $qlims($keyvals(queue)) $idx(qmaxnodes)]
	    set keyvals(ppn) [lindex $qlims($keyvals(queue)) $idx(qppn)]

	    # Reset list of available resources. 

	    set w [lindex $keywidgets(resource) 0]
	    $w configure -values $qresources($keyvals(queue))
	    wqCheck resource
	    wqCheck nodes
	    set rtn 1
	}

	resource {

	    # Makes sure resource is in the list of available resources
	    # for the selected queue. If not, set to first item in list.
	    # Should only happen when changing queues.

	    if { [lsearch -exact $qresources($keyvals(queue)) \
		      $keyvals(resource)] == -1 } {

		set w [lindex $keywidgets(resource) 0]
		$w set [lindex $qresources($keyvals(queue)) 0]

	    }

	    set rtn 1
	}

	nodes {

	    # Clip to maximum node count.

	    if { $keyvals(nodes) > \
		     [lindex $qlims($keyvals(queue)) $idx(qmaxnodes)] } {

		set keyvals(nodes) [lindex $qlims($keyvals(queue)) \
					$idx(qmaxnodes)]
	    }

	    # Make sure node count isn't negative.

	    if { $keyvals(nodes) < 1 } {

		tk_messageBox -title "NODES < 1" -icon info -type ok \
		    -message "Node value < 1 - Reset to 1"
		set keyvals(nodes) 1

	    }

	    set rtn 1
	}

	priority {

	    if { $keyvals(priority) < -1024 } { 
		tk_messageBox -title "PRIORITY < -1024" -icon info -type ok \
		    -message "Resetting priority value to -1024"

		set keyvals(priority) -1024

	    }

	    if { $keyvals(priority) > 1024 } {
		tk_messageBox -title "PRIORITY > 1024" -icon info -type ok \
		    -message "Resetting priority value to 1024"

		set keyvals(priority) 1024

	    }

	    set rtn 1
	}

	chkpt_interval {

	    # This is the most convoluted option, as it can be
	    # switches, but one may take an optional value as well.

	    set val $keyvals(chkpt_interval)
	    set char [lindex $keyvals(chkpt_interval) 0]
	    set ok 0

	    switch $char {

		"" {
		    set ok 1
		}

		n {
		    if { [string length $val] == 1 } set ok 1
		}

		s {
		    if { [string length $val] == 1 } set ok 1
		}

		c {
		    if { [string length $val] == 1 } {

			set ok 1

		    } else {

			set tmp [split $val =]

			if { [llength $tmp] == 2 && \
				 [string is integer [lindex $tmp 1]] } {

			    if { $tmp > 0 } {

				set ok 1
			    }
			}
		    }
		}
		default {
		    set ok 0
		}
	    }

	    if { $ok == 0 } { 

		tk_messageBox -title "BAD CHKPT INTERVAL" -icon info \
		    -type ok \
		    -message "Bad setting: \"$keyvals(chkpt_interval)\""

		set keyvals(chkpt_interval) $keydefaultvals(chkpt_interval)
	    }

	    set rtn 1
	}

	workdir {

	    if { [string length $keyvals(workdir)] == 0 } {

		set rtn 0

	    } else {

		set rtn 1

            }
	}

	infiles {
	    if { [string length $keyvals(infiles)] == 0 } {

		set rtn 0

	    } else {

		set rtn 1

	    }
	}

	default {

	    # Debug "No validation set up for $key"

	    set rtn 1
	}
    }

    return $rtn
}


proc wqEntry { key e } {

    # Creates an entry control with label and entry widget.

    global keyword
    global idx
    global keydefaultvals
    global keywidgets
    global keyvals

    set k $keyword($key)
    set keyvals($key) ""
    frame $e
    label $e.label -text "[lindex $k $idx(keyname)]" -width 20 -anchor e
    pack $e.label -side left
    set d $keydefaultvals($key)

    if { [string length $d] == 0 } {

	set d "none"

    }

    balloonhelp_for $e.label [lindex $k $idx(keytext)]
    set keyvals($key) $keydefaultvals($key)
    ttk::entry $e.info -textvariable keyvals($key)
    pack $e.info -fill x -padx 2
    set keywidgets($key) [list $e.info $e.label]
    pack $e -side top -fill x
    bind $e <Leave> [list wqCheck $key]
}


proc wqDirSelect { key e } {

    # Creates a label and control to allow selection of a directory.

    global keyword
    global idx
    global keydefaultvals
    global keywidgets
    global keyvals

    set k $keyword($key)
    set keyvals($key) ""
    frame $e
    label $e.label -text "[lindex $k $idx(keyname)]" -width 20 -anchor e
    pack $e.label -side left
    balloonhelp_for $e.label [lindex $k $idx(keytext)]
    ttk::entry $e.name -textvariable keyvals($key) -width 20 -state readonly
    pack $e.name -padx 2 -side left

    if { [string length $keydefaultvals($key)] > 0 } {

	set keyvals($key) $keydefaultvals($key)

    }

    ttk::button $e.sel -text "..." -width 2 \
	-command [list wqOpen $key d ] 
    pack $e.sel
    set keywidgets($key) [list $e.name $e.label]
    pack $e -side top -fill x
    bind $e <Leave> [list wqCheck $key]
}


proc wqFileSelect { key e } {

    # Create a label and control to support file selection.

    global keyword
    global idx
    global keydefaultvals
    global keywidgets
    global keyvals

    set k $keyword($key)
    set keyvals($key) ""
    frame $e
    label $e.label -text "[lindex $k $idx(keyname)]" -width 20 -anchor e
    pack $e.label -side left
    balloonhelp_for $e.label [lindex $k $idx(keytext)]
    ttk::entry $e.name -textvariable keyvals($key) -width 20 -state readonly
    pack $e.name -padx 2 -side left

    if { [string length $keydefaultvals($key)] > 0 } {

	set keyvals($key) $keydefaultvals($key)

    }

    ttk::button $e.sel -text "..." -width 2 \
	-command [list wqOpen $key f ] 
    pack $e.sel
    set keywidgets($key) [list $e.name $e.label]
    pack $e -side top -fill x
    bind $e <Leave> [list wqCheck $key]
}


proc wqSpinBox { key e } {

    # Create a label and spinbox control for value selection.

    global keyword
    global idx
    global keywidgets
    global keyvals

    set k $keyword($key)
    set opts [lindex [lindex $k $idx(keydefault)] 0]
    set default [lindex [lindex $k $idx(keydefault)] 1]
    set keyvals($key) ""

    frame $e
    label $e.label -text "[lindex $k $idx(keyname)]" -width 20 -anchor e
    pack $e.label -side left
    balloonhelp_for $e.label "[lindex $k $idx(keytext)]\n Default: $default"
    ttk::spinbox $e.spinner -from [lindex $opts 0] -to [lindex $opts 1] \
	-increment 1 -format "%5.0f" -textvariable keyvals($key)
    $e.spinner set $default
    set keywidgets($key) [list $e.spinner $e.label]
    pack $e.spinner -padx 2 -fill x
    pack $e -side top -fill x
    bind $e <Leave> [list wqCheck $key]
}


proc wqComboBox { key e } {

    # Create a label and a drop-down combo box for value selection.

    global keyword
    global idx
    global keywidgets
    global keyvals

    set k $keyword($key)
    set opts [lindex [lindex $k $idx(keydefault)] 0]
    set default [lindex [lindex $k $idx(keydefault)] 1]
    set keyvals($key) ""

    frame $e
    label $e.label -text "[lindex $k $idx(keyname)]" -width 20 -anchor e
    pack $e.label -side left
    balloonhelp_for $e.label "[lindex $k $idx(keytext)]. Default: $default"
    ttk::combobox $e.combobox -state readonly -values $opts \
	-textvariable keyvals($key)
    $e.combobox set $default
    set keywidgets($key) [list $e.combobox $e.label]
    pack $e.combobox -padx 2 -fill x
    pack $e -side top -fill x
    bind $e <Leave> [list wqCheck $key]
}


proc wqOpen { key type } {

    # Support selection of a file or directory.

    global keyvals

    set old $keyvals($key)

    switch $type {

	d { set ans [tk_chooseDirectory] }

	f { set ans [tk_getOpenFile] }

    }

    if { [string length "$ans"] > 0 } {

	set keyvals($key) $ans

    }
}


proc wqGenerate { } {

    # Placeholder. Will eventually generate PBS and other files.

    global keyword
    global keyvals
    global pbsTemplate
    global idx

    # Make some checks not done elsewhere.

    if { [wqCheck infiles] == 0 } {

	tk_messageBox -title "MISSING KEYWORD" -icon warning -type ok \
	    -message "[lindex $keywords(infiles) $idx(keyname)] must be set."
	return
    }

    if { [string length $keyvals(resource)] > 0 } {

	set rsrc ":$keyvals(resource)"

    } else {

	set rsrc ""
    }

    set msg [format $pbsTemplate(head) $keyvals(allocation) \
		 $keyvals(nodes) $keyvals(ppn) $rsrc $keyvals(walltime) \
		 $keyvals(queue) $keyvals(job_name)]

    # Process all the options which default to "n":

    foreach key [list email_options join_list keep_list full_env restartable \
		     suppress_jobid] {

	if { [string equal $keyvals($key) n] == 0 } {

	    set s [lindex $keyword($key) $idx(keyswitch)]
	    append msg [format "#PBS %s %s\n" $s $keyvals($key)]

	}
    }

    # Process all the options which default to blank:

    foreach key [list email_address stdout_path stderr_path env_list \
		     datetime chkpt_interval shell_path_list] {

	set opt [string trim $keyvals($key)]

	if { [string length $opt] > 0 } {

	    set s [lindex $keyword($key) $idx(keyswitch)]
	    append msg [format "#PBS %s %s\n" $s $opt]

	}
    }

    if { $keyvals(priority) != 0 } {

	set s [lindex $keyword(priority) $idx(keyswitch)]
	append msg [format "#PBS %s %d\n" $s $keyvals(priority)]

    } 

    set wpn [expr $keyvals(ppn) / $keyvals(procs_per_task)]
    append msg [format $pbsTemplate(prologue) $wpn $keyvals(workdir) \
		    $keyvals(infiles) $keyvals(pbsfile) $keyvals(start) \
		    $keyvals(task) ]
    append msg $pbsTemplate(epilogue)

    if { [file exist $keyvals(pbsfile)] } {

	set ans [tk_messageBox -title "OVERWRITE?" -icon warning -type yesno \
		     -message "Overwrite PBS file: \"$keyvals(pbsfile)\"?"]

	if { [string compare $ans no] == 0 } { return }

    }

    if [catch {open $keyvals(pbsfile) w 0644} stream] {

	tk_messageBox -title "FILE OPEN FAILURE" -icon error -type ok \
		-message "Failed to PBS script file: \"$keyvals(pbsfile)\""

    } else {

	puts $stream $msg

	tk_messageBox -title "FILE GENERATED" -icon info -type ok \
		-message "Generated PBS script file: \"$keyvals(pbsfile)\""
    }

    close $stream
    # tk_messageBox -icon info -type ok -message $msg
}


proc wqWindow { } {

    # Create the main dialog box and all controls.

    global bold
    global host
    global keyword
    global keywords
    global keydefaultvals
    global idx
    global entry_field
    global spin_box
    global combo_box

    # Creates the WQ window.
    
    # Set file info line.

    frame .fileInfo

    # Define the File menu button.

    set b .fileInfo.wqFileMenuButton

    menubutton $b -background bisque -borderwidth 4 -menu "$b.m" \
	-relief raised -text File -font $bold

    menu $b.m

    $b.m add command -label "New" -command { wqNew }
    $b.m add command -label "Open .." -command { wqOpen } 
    $b.m add command -label "Save .." -command { wqSave }
    $b.m add separator
    $b.m add command -label "Revert" -command { wqRevert }
    $b.m add separator
    $b.m add command -label "Generate" -command { wqGenerate }
    $b.m add separator
    $b.m add command -label "Quit" -command { wqQuit }

    label .fileInfo.wqActiveFileLabel -background white -foreground black \
	-justify left -padx 6 -pady 6 -relief groove -text label -font $bold \
	-textvariable wqFile

    pack $b -side left -anchor w
    pack .fileInfo.wqActiveFileLabel -fill x -side left 
    pack .fileInfo -side top -fill x

    frame .sep1 -height 2 -borderwidth 1 -relief sunken
    pack .sep1 -side top -fill x -padx 4 -pady 4

    # Create the notebook tab sections.

    tabnotebook_create .tn
    pack .tn
    set pbs_page [tabnotebook_page .tn "PBS"]
    set task_page [tabnotebook_page .tn "Task"]

    # Create content for the PBS page

    label $pbs_page.title -text "Enter PBS job information for $host:"
    pack $pbs_page.title -anchor w

    frame $pbs_page.sep2 -height 2 -borderwidth 1 -relief sunken
    pack $pbs_page.sep2 -side top -fill x -padx 4 -pady 4

    scrollform_create $pbs_page.sform
    pack $pbs_page.sform -expand yes -fill both

    set form [scrollform_interior $pbs_page.sform]

    label $form.required -text "Required Keywords:" -font $bold -fg black
    pack $form.required -anchor w
    set counter 0

    foreach key $keywords {

	set k $keyword($key)

	if { [lindex $k $idx(keytag)] == $idx(required)} {

	    set line "$form.reqline[incr counter]"

	    if { [string length [lindex $k $idx(keydefault)]] != 0 && \
		 [string length $keydefaultvals($key)] == 0 } {

		set keydefaultvals($key) [lindex $k $idx(keydefault)]
	    }

	    switch [lindex $k $idx(keytype)] \
		$idx(file_select) { wqFileSelect $key $line } \
		$idx(dir_select) { wqDirSelect $key $line } \
		$idx(entry_field) { wqEntry $key $line } \
		$idx(spin_box) { wqSpinBox $key $line } \
		$idx(combo_box) { wqComboBox $key $line } 
	}
    }

    set line "$form.reqline[incr counter]"
    frame $line -height 2 -borderwidth 1 -relief sunken
    pack $line -fill x -padx 4 -pady 4

    label $form.optional -text "Optional Keywords:" -font $bold -fg blue
    pack $form.optional -anchor w
    set counter 0

    foreach key $keywords {

	set k $keyword($key)

	if { [lindex $k $idx(keytag)] == $idx(optional)} {

	    set line "$form.optline[incr counter]"

	    if { [string length [lindex $k $idx(keydefault)]] != 0 && \
		 [string length $keydefaultvals($key)] == 0 } {

		set keydefaultvals($key) [lindex $k $idx(keydefault)]
	    }

	    switch [lindex $k $idx(keytype)] \
		$idx(file_select) { wqFileSelect $key $line } \
		$idx(dir_select) { wqDirSelect $key $line } \
		$idx(entry_field) { wqEntry $key $line } \
		$idx(spin_box) { wqSpinBox $key $line } \
		$idx(combo_box) { wqComboBox $key $line } 
	}
    }

    set line "$form.optline[incr counter]"
    frame $line -height 2 -borderwidth 1 -relief sunken
    pack $line -fill x -padx 4 -pady 4

    label $form.advanced -text "Advanced Keywords:" -font $bold -fg red
    pack $form.advanced -anchor w
    set counter 0

    foreach key $keywords {

	set k $keyword($key)

	if { [lindex $k $idx(keytag)] == $idx(advanced)} {

	    set line "$form.exoline[incr counter]"

	    if { [string length [lindex $k $idx(keydefault)]] != 0 && \
		 [string length $keydefaultvals($key)] == 0 } {

		set keydefaultvals($key) [lindex $k $idx(keydefault)]
	    }

	    switch [lindex $k $idx(keytype)] \
		$idx(file_select) { wqFileSelect $key $line } \
		$idx(dir_select) { wqDirSelect $key $line } \
		$idx(entry_field) { wqEntry $key $line } \
		$idx(spin_box) { wqSpinBox $key $line } \
		$idx(combo_box) { wqComboBox $key $line } 
	}
    }

    # Create content for the Task page. This will be an editing window
    # to allow creation of a task script.

    label $task_page.title -text "Enter task script below:"
    pack $task_page.title -anchor w

    frame $task_page.sep2 -height 2 -borderwidth 1 -relief sunken
    pack $task_page.sep2 -side top -fill x -padx 4 -pady 4

    # Create the edit window.

    set form $task_page.editor
    frame $form
    scrollbar $form.sbar -command [list $form.text yview]
    text $form.text -yscrollcommand [list $form.sbar set]
    pack $form.sbar -side right -fill y
    pack $form.text -side left -expand yes -fill both
    pack $form -expand yes -fill both -padx 4 -pady 4
}


proc wqReadConfig { } {

    # Read in any .wqrc file that might exist, and initialize some
    # variables. Where we look for the config file will depend on what
    # OS is being used.

    global keywords
    global keydefaultvals
    global tcl_platform
    global env

    # Initialize all default values to null.

    foreach k $keywords {

	set keydefaultvals($k) ""
    }

    # Next, determine where the file might be. For Windows, there's
    # only one place to look since it has such a poor sense of working
    # directory when wish is started - the user profile. For Linux,
    # check the current working directory first, then the user's home
    # directory.

    set rcFile ""

    if { [string compare $tcl_platform(platform) windows] == 0 } {

	set f "$env(USERPROFILE)\\.wqrc"
	set keydefaultvals(user_name) $env(USERNAME)

	if [file exists $f] {

	    set rcFile $f
	}

    } else {

	set keydefaultvals(user_name) $env(USER)

	if [file exists .wqrc] {

	    set rcFile .wqrc

	} else {

	    set f "$env(HOME)/.wqrc"

	    if [file exists $f] {
		set rcFile $f
	    }
	}
    }
	    
    # If a .wqrc file exists, read in the configuration settings.

    if { [string length $rcFile] > 0 } {

	if [catch {open $rcFile r} stream] {

	    tk_messageBox -title "FILE OPEN FAILURE" -icon error -type ok \
		-message "The file \"$rcFile\" can't be read."

	    return
	}

	# Read everything in, splitting out individual lines.  Then
	# scan each line, identify the keyword, and use the value to
	# initialize the appropriate widget. Lines begining with # are
	# treated as comments.

	foreach line [split [read $stream] \n] {

	    set v {}
	    set k ""

	    if { [string length "$line"] != 0 } {

		foreach f [split $line] {

		    if { [string length "$k"] == 0 } {

			set k [string tolower $f]

		    } else {

			if { [string length "$f"] != 0
			     && [string compare "$f" "="] != 0 } {

			    lappend v "$f"
			}
		    }
		}
	    }

	    if { [string length "$k"] > 0 && [string first "#" $k] != 0 } {

		if { [lsearch -exact $keywords $k] != -1 } {

		    set keydefaultvals($k) "$v"

		} else {

		    tk_messageBox -title "UNKNOWN KEYWORD" -icon error \
			-type ok -message "Culprit in .wqrc file:  \"$k\""
		}
	    }
	}

	close $stream
    }
}


# Processing actually starts below. The variable host must be set
# to the target machine in order to configure things properly.

global host
set host philip

wqReadConfig
wqSettings
wqWindow

# Check the starting keyword values to make sure they are sane.

foreach key $keywords {

    wqCheck $key
}

# That's it. The window will run until exited.
