#! /bin/bash
########################################################################
# Begin WQ PBS support script.
#
# This script is closely coupled with the behavior and proper usage of
# wq.py, and provides the interface to the PBS batch system. It should
# be considered powerful mojo and to be dabbled with at your own
# peril.
#
# Copyright (C) 2014-2015  James A. Lupo
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Revision: $Id: wq-pbs.sh 246 2015-12-11 19:44:28Z jalupo $
#
# The script wq.py is expected to be executable and in the jobs PATH.
#
# While not meant to be called directly by the user, this is how the
# PBS script calls it with the various user settings:
#
#    wq-pbs.sh CMD WPN WORKDIR FILES START TASK VERBOSE JOBNUM
#
#    CMD ...... Is the name of the PBS job script.
#    WPN ...... Number of workers desired per node.
#    WORKDIR .. The working directory to use.
#    FILES .... The name of the file holding the input list.
#    START .... The starting record number in FILES.
#    TASK ..... The path to the task command.
#    VERBOSE .. Set how chatty WQ should be.
#    JOBNUM ... The PBS job number - blank on mother superior,
#               an integer string on compute nodes.    
#
########################################################################

# Check number of arguments, then assign to named variables:

if [[ $# -lt 7 ]]; then
   echo "WQ-PBS.SH: Wrong number of arguments."
   echo 'Usage:  wq-pbs.sh CMD WPN WORKDIR FILES START TASK VERBOSE JOBNUM'
   echo '        JOBNUM should be blank when called from mother superior.'
   exit 1
fi

JOBFILENM=$1
WPN=$2
WORKDIR=$3
FILES=$4
START=$5
TASK=$6
VERBOSE=$7
JOBNUM=$8

# Set verbosity flag, if desired.

if [ "${VERBOSE}" = "1" ] ; then
   VERBOSE=--verbose
else
    VERBOSE=
fi

# Drop into the working directory after making sure it exists.

if [ ! -d ${WORKDIR} ] ; then
   echo "WQ.PBS Error: WORKDIR = \"${WORKDIR}\" does not exist!"
   exit 1
fi

cd ${WORKDIR}

# Only the mother superior has PBS_JOBID defined, so we will be
# using that fact to pass the job number to the other nodes as the
# only command line argument. It becomes the last argument in the
# call this for this script, thus, if the variable JOBNUM is blank,
# we must be running on the mother superior. Otherwise, we're on a
# compute node:

if [ "${JOBNUM}x" = "x" ] ; then

   # Must be running on the mother superior. Do some basic sanity
   # checking just to be safe.

   if [ ! -r ${FILES} ] ; then
      echo "WQ.PBS Error: FILES = \"${FILES}\" does not exist or can't be read!"
      exit 1
   fi

   if [ $(wc -l ${FILES} | cut -d ' ' -f 1) -lt 1 ] ; then
      echo "WQ.PBS Warning: FILES = \"${FILES}\" is empty. No work to do!"
      exit 0
   fi

   if [ ! -x ${TASK} ] ; then
      echo "WQ.PBS Error: TASK = \"${TASK}\" does not exist or isn't executable!"
      exit 2
   fi

   if [ ${START} -lt 1 ] ; then
      echo "WQ.PBS Error: START can't be less than 1! Quiting!"
      exit 3
   fi

   # Make sure the dispatcher port is available.
   
   PORT=54321
   P=`netstat -lt | gawk 'BEGIN{x=0;}/'$PORT'/{x=1;}END{print x;}'`
   if [ $P = '1' ] ; then
       echo "WQ.PBS Error: Port $PORT is busy?!? Quiting!"
       exit 4
   fi 

   # Remember our host name.

   MS=`uname -n`

   # Use a bit of magic to strip off the trailing host name and
   # leave only the job number from PBS_JOBID:

   JOBNUM=${PBS_JOBID%.*}
   HOSTLIST=${WORKDIR}/hostlist.${JOBNUM}

   # We want the mother superior host name first. So, take the host
   # list provided, sort it into a unique list of names, with MS first.
   # This assures it's node ID, or position in the hostlist, is 1.

   echo ${MS} > ${HOSTLIST}
   grep -v ${MS} ${PBS_NODEFILE} | uniq | sort >> ${HOSTLIST}

   # Compute the number of nodes assigned.

   export NODES=`wc -l ${HOSTLIST} |gawk '//{print $1}'`
   
   # Make a local copy of the PBS script since only the mother superior
   # can see it at job start.

   JOBFILE=${WORKDIR}/pbs.${JOBNUM}
   cp $JOBFILENM $JOBFILE
   chmod u+x ${JOBFILE}

   # Mother superior must start up the dispatcher, so:

   wq.py --dispatcher ${TASK} --start $START \
       $VERBOSE --time ${PBS_WALLTIME} --inputs ${FILES} \
       --allworkers $(( $WPN * $NODES )) --jobid ${JOBNUM} &

   # Give it a chance to spin up since it would be nice to have
   # the dispatcher ready to accept connections from the workers
   # as they start. Not strictly necessary with zmq sockets, but
   # old habits die hard.

   sleep 5

   # Ready to start the script on all compute nodes. This will fire
   # up workers. We'll pass the job number as the only argument.
   # They'll connect to the dispatcher and start work immediately.

   for H in `cat ${HOSTLIST}` ; do
      if [ ${H} != ${MS} ] ; then
         ssh -n ${H} ${JOBFILE} ${JOBNUM} &
      fi
   done

   # Finally, mother superior can start it's own workers:

   wq.py --workers ${WPN} --mothersuperior ${MS}

   # Make sure to wait until all the processes are done!

   wait

else

   # Must be running on a compute node. The job number was passed by the
   # mother superior (see above).

   HOSTLIST=${WORKDIR}/hostlist.${JOBNUM}

   # Now, we have to get the name of mother superior from the host
   # list. Thats so we know where the dispatcher is running. Simply
   # grab the first entry from the hostlist file and press on.

   MS=`head -1 ${HOSTLIST}`

   # Ready to go. Spin up the workers. The mother superior passed
   # the job wall time as argument 1 when the script is called, so
   # we have all the values needed for workers:

   wq.py --workers ${WPN} --mothersuperior ${MS}

fi

# Give a bit of time to make sure dispatcher has shut down cleanly.
# The WAIT above should allow for this, but coming down on the side of
# paranoia:

sleep 2
