#! /bin/env python
"""WQ (WorkQueue) distributed computing controller based on zmq.

    Copyright (C) 2014-2015  James A. Lupo

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    jalupo@cct.lsu.edu  or  jalupo2009@gmail.com


There are 2 components provided by this script: 

   dispatcher ...... maintains queue of tasks, provides task data on
                     request, tracks remaining job time and initiates
                     shutdown when there is no more work, or estimates
                     insufficient job time remains to complete tasks.

   worker .......... continuously requests tasks from the dispatcher,
                     one at a time, until all tasks are completed or
                     is told to stop.

In a parallel cluster environment, the dispatcher must be the first
program launched on the mother superior. Workers should then be
launched on all nodes, including the mother superior.

Work does not begin until the dispatcher receives and responds to a
request for a task assignment. Work will be handed out until the list
of tasks is exhausted or job time runs out.

The most important configuration occurs in the main line below. They
are:

  port ..... Port for the dispatcher request socket (54321)
  jobtime .. Default total wallclock time in seconds
             (86400 seconds [1 day])
"""

import time
import zmq
import getopt
import sys
import socket
import subprocess
from multiprocessing import Process
import re
import os

# ZMQ is the only required module which might be missing:

try:
   import zmq
except:
   print "Failed to import module zmq!"
   print "Add module, or try different Python version/distribution."
   print "Currently running: ", sys.version
   sys.exit(1)


def shell ( cmd ):
   """
   Submits a shell command for processing, and returns the cmd status
   flag plus the STDIN and STDOUT messages.

   Arguments:

      cmd .. The desired shell command line.

   Returns:

      3-tuple: True/False for success/failure.
               List of stdout lines.
               List of stderr lines.
   """
   p = subprocess.Popen( cmd, shell=True, stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE )
   x = p.communicate()
   p.stdout.close()
   p.stderr.close()
   if x[1] == '':
      status = True
   else:
      status = False
      
   return [ status, x[0].split( '\n' ), x[1].split( '\n' ) ]


def ipaddrs( host ):
   """
   Gets IP address for a host specified by name.

   Arguments:
    
      host .. Host name to look up.

   Returns:

      The named host's IP address.
   """
   return socket.gethostbyaddr(host)[2][0]


def physcpubindings( workers ) :
    """
    Determines how many cores are present, then creates the numactl
    command to assign a unique set of cores to each worker. If needed
    by the user script, it would be accessed from the environment
    variable WQ_NUMACTL and applied as so:

       eval "$WQ_NUMACTL <user command line>"

    It also sets a string for WQ_CPT (cores per task), which can be
    used to get the number of cores to use for MPI processes or
    OpenMP threads.

    The return value is a list with:
    
       [0] : WQ_NUMACTL
       [1] : WQ_CPT

    """
    cmd = shell( 'numactl --show' )

    # Extract the 'physcpubind' entry, if we can find one. The output
    # may vary, especially if NUMA polices aren't supported.
    
    physcpubind = False
    
    if cmd[0]:
        for line in cmd[1]:
            if line.find("physcpubind:") == 0:
                physcpubind = True
                break
             
    if physcpubind:
        cpu_list = line.split(':')[1].strip().split(' ') 
        ppn = len(cpu_list)
        ppw = ppn / workers
        cpt = 'WQ_CPT=%d'%(ppw)
        i = 0
        bindings = []
        for w in range(workers) :
           s = '%d-%d'%( w * ppw, (w + 1) * ppw - 1 )
           bindings.append( 'WQ_NUMACTL="numactl --physcpubind=%s -- "'
                            % ( s ) )
    else:
       # Might get here if the machine doesn't support NUMA.\
       cpt = 'WQ_CPT=1'
       bindings = [ 'WQ_NUMACTL=" "' ]

    return bindings,cpt


def dispatcher( port, cmd, lines, allworkers, start, jobtime, 
                verbose, jobid ):
   """
   The dispatcher task, which is run as a separate thread, handles
   distribution of tasks to workers. Workers must request a task
   then wait for data to be sent in reply.

   Arguments:

      port ......... The socket on which to listen for work requests.
      cmd .......... Command for workers to execute. It will be provide
                     1 line as it's arguments.
      lines ........ List of input lines to distribute.
      allworkers ... Total number of workers (workers per node * nodes).
      start ........ The task number to start with - allows skipping
                     over previously completed tasks.
      jobtime ...... Walltime, in seconds, available.
      verbose ...... Be chatty about activity if True.
      jobid ........ Job ID (number) string.

   The dispatcher reads a list of lines from an input file and
   constructs task messages for the workers. Workers must first issue
   a request to the dispatcher, which then replies with one task
   message.  When all work is handed out, or it estimates that
   insufficient time remains to start new tasks, the dispatcher starts
   sending termination messages instead until all workers have been
   notified to cease.

   The work request message is a dictionary with one of two forms. For the
   initial startup request, it looks like:

      'mode' : 'start' ........... Indicates start request.
      'worker' : workerID ........ Worker name (hostnm_#)
      'localtime' : time.time() .. Start time local to worker.
      'affinity' : bindings ...... NUMACTL setting.
      'cpt' : cpt ................ Cores per thread.

   For all follow on requests, it looks like:
 
      'mode' : 'ran' ............. Indicates task was run.
      'worker' : workerID ........ Worker name (hostnm_#)
      'tasknum' : tasknum ........ The task number completed.
      'task' : task .............. The full task command line.
      'status' : result[0] ....... True for success, False for error.
      'stdout' : result[1] ....... All task stdout lines.
      'stderr' : result[2] ....... All task stderr lines.
      'tasktime' : elapsed ....... Task run time (seconds).
      'taskstart' : taskstart .... Worker local time of task start.
      'taskend' : taskend ........ Worker local time of task end.

   The tasking response message is a dictionary of:
       'cmd' ...... The command to execute, or "FINI" to quit.
       'line' ..... The next line (cmd) to be processed.
       'tasknum' .. The sequence number of the assigned task.

   """

   # Open a log file. Any other output goes to stderr.

   logname = 'wq.log'
   if jobid != '' :
      logname = logname + '.' + jobid

   log = open( logname, 'w+' )
   
   # Set a starting time hack, then get the name of the host we're
   # running on, and create a ZMQ context for communication.

   starttime = time.time()

   # Set up a socket to receive task requests and send replies over.
   # The linger option is set to help make sure all comunication is
   # delivered when the thread ends. The time unit is milliseconds.
   # A rigorous 'receive request' - 'send reply' pattern must be
   # followed as the zmq.REP socket keeps track of who sent the
   # request and thus were the reply should go to. Trying to do two
   # receives or two sends in a row will cause a fatal error or hang
   # the program. Here we set up the REP side of the socket pattern.

   host = ipaddrs( socket.gethostname() )
   context = zmq.Context()
   dispatcher_socket = context.socket( zmq.REP )
   dispatcher_socket.setsockopt( zmq.LINGER, 5000 )

   try:
      dispatcher_socket.bind( "tcp://%s:%s" % ( host, port ) )
   except zmq.ZMQError, err:
      sys.stderr.write( "Dispatcher: Failed to bind to port %s\n" \
                        % ( port ) )
      sys.exit ( 2 )

   # Initialize some of the things we need to keep track of.
   
   maxtime = 0.0         # Estimate of time required for a task.
   tasknum = 0           # Sequence number of assigned task.
   margin = 1.25         # Safety margin: maxtime = margin * taskmaximum
   taskcount = 0         # Count of tasks completed.
   taskaverage = 0.0     # Accumulator for average task time stat.
   taskminimum = 1.e+36  # Captures shortest task execution time.
   taskmaximum = 0.0     # Captures longest task execution time.
   runtimeave = 0.0      # Average worker total run time.
   runtimemin = 0.0      # Run time of first worker to terminate.
   runtimemax = 0.0      # Run time of last worker to terminate

   # This dictionary will map worker names to run status: 0 - terminated,
   # 1 - working.

   workers = {}

   # And we'll also count how many workers have been told we're done.

   already_notified = 0

   log.write( "Dispatcher:Start:%d:%.2f\n" % ( start, starttime ) )
   log.flush()

   # Adjust starting task for 0 offset:

   start = start - 1
   tasknum = start

   for l in lines[start:]:

      request = dispatcher_socket.recv_json()
      worker = request['worker']

      # Keep track of elapsed job time and job time remaining.
      
      walltime = time.time() - starttime
      timeleft = jobtime - walltime

      print_results( log, request, verbose )
      
      if request['mode'] == 'start' :
         log.write( 'Worker:Starting:%s:%.2f\n' % (worker,walltime) )
      elif request['mode'] == 'ran' :
         taskcount += 1
         taskaverage += request['tasktime']
         if request['tasktime'] < taskminimum :
            taskminimum = request['tasktime']
         if request['tasktime'] > taskmaximum :
            taskmaximum = request['tasktime']
            maxtime = taskmaximum * margin
            log.write( "Dispatcher:Maxtime:%s:%.2f:%.2f\n"
                       % ( worker, maxtime, time.time() ) )
            log.flush()

      if timeleft >= maxtime :

         line = l.strip()

         if verbose :
            log.write( "Dispatcher:Task:%d:%s:%s\n"
                       % ( tasknum + 1, worker, line ) )
            log.flush()

         tasknum = tasknum + 1
         workers[worker] = 1
         task_message = { 'cmd' : cmd,
                          'line' : line,
                          'tasknum' : tasknum }

      else:

         workers[worker] = 0
         log.write( "Dispatcher:Assigning:%s:-1:FINI\n" % ( worker ) )
         log.flush()
         task_message = { 'cmd' : "FINI",
                          'line' : "None",
                          'tasknum' : tasknum }
         already_notified += 1
         runtimemin = time.time() - starttime
         runtimemax = runtimemin
         runtimeave += runtimemin

      dispatcher_socket.send_json( task_message )
      if workers[worker] < 1 :
         break

   # Now make sure all workers have received the shutdown message.

   shutdown = allworkers - already_notified
   
   # Could we scan workers[] and report who has not yet completed?
   # Maybe time how long we have been waiting and maybe detect a
   # a dead or hung worker?
   
   if shutdown > 0 :

      if verbose :
         log.write( "Dispatcher:WaitOn:%d\n" % ( shutdown ) )
         log.flush()
         i = 0
         msg = "Dispatcher:WaitFor"
         for h in workers:
            msg = msg + ":%s" % ( h )
            i = i + 1
            if i%6 == 0:
               log.write( "%s\n" % ( msg ) )
               log.flush()
               if i < shutdown:
                  msg = "WaitFor"
         if i%6 != 0:
            log.write( "%s\n" % ( msg ) )
            log.flush()
      
      task_message = { 'cmd' : "FINI", 'line' : "None",
                       'maxtime' : -1, 'tasknum' : tasknum }

      # Now wait for everyone to finish up, grabbing the final task info.

      for w in range( shutdown ):
         request = dispatcher_socket.recv_json()
         print_results( log, request, verbose )
         dispatcher_socket.send_json( task_message )
         runtime = time.time() - starttime
         if runtimeave == 0.0 :
            runtimemin = runtime
            runtimemax = runtimemin
         if runtime > runtimemax :
            runtimemax = runtime
         runtimeave += runtimemax
         taskcount += 1
         taskaverage += request['tasktime']
         if request['tasktime'] < taskminimum :
            taskminimum = request['tasktime']
         if request['tasktime'] > taskmaximum :
            taskmaximum = request['tasktime']

   log.write( "Dispatcher:Last:%d\n" % ( tasknum )
              + "Dispatcher:Shutdown:%d:%.2f:%.2f:%.2f:%.2f:%.2f:%.2f\n"
              % ( taskcount, taskaverage/taskcount, taskminimum,
                  taskmaximum, runtimeave/allworkers, runtimemin,
                  runtimemax ) )
   log.flush()
   dispatcher_socket.setsockopt( zmq.LINGER, 0 )
   dispatcher_socket.close()
   context.term()


def worker( wrk_num, host, port, verbose, bindings, cpt ):
   """
   Defines the worker task. The arguments include:

      wrk_num .. Identifier for the worker on a node.
      host ..... IP of host running the dispatcher.
      port ..... The dispatcher port.
      verbose .. Be chatty about activity if True.
      bindings . The WQ_NUMACTL variable string that holds the
                 processor affinity command assigning unique cores
                 to this worker as set by physcpubindings().
      cpt ...... The WQ_CPT variable string that holds the number
                 of cores per task.

   The "worker" sends a task request message via a zmq.REQ socket to
   the dispatcher, and waits for a reply. Each reply is a dictionary
   containing a command name and an input file. If the command is not
   the termination command, the task is executed and the result is sent
   down  zmq.PUSH connection to the results manager.

   A "task" is defined in a dictionary set by the dispatcher, and the
   results are returned in a second dictionary.  Obviously, the keys
   must agree on both ends. See dispatcher above for a description
   of the request and reply messages.
   """
   # Get a starting time hack (in seconds since the epoc).

   starttime = time.time()

   # Get our host name and create a worker name:

   local = socket.gethostname()
   workerID = "%s_%d" % ( local, wrk_num )

   # Initialize a zeromq context

   context = zmq.Context()

   # Set up a socket for communication with the dispatcher. This
   # is the requestor side of the REQ/REP pattern.

   try:
      task_socket = context.socket( zmq.REQ )
      task_socket.connect( "tcp://%s:%s" % ( host, port ) )
      task_poller = zmq.Poller()
      task_poller.register( task_socket, zmq.POLLIN )
   except zmq.ZMQError, err:
      sys.stderr.write( "ERROR: Worker %s failed to create REQ socket.\n" \
                        % ( workerID ) )
      # No sane way to recover if we can't talk to dispatcher!
      sys.exit(1)
   
   running = True
   tasknum = 0
   maxtime = 0.0

   # Wait no more than 30 seconds for a response from the dispatcher.
   # The time here is expressed in milliseconds.

   timeout = 30000.0

   # Send the initial request for a task.

   task_socket.send_json( { 'mode' : 'start',
                            'worker' : workerID,
                            'localtime' : time.time(),
                            'affinity' : bindings,
                            'cpt' : cpt } )
                            
   while running:

      socks = dict( task_poller.poll( timeout ) )

      if socks.get( task_socket ) == zmq.POLLIN:

         # Looks like we received a task. Process it.

         task_message = task_socket.recv_json()

         if task_message['cmd'] != "FINI" :

            tasknum = task_message['tasknum']

            # Construct the command line.

            task = "%s %s %s %s" % ( cpt, bindings, task_message['cmd'],
                                  task_message['line'] )
            taskstart = time.time()

            # Do it!

            result = shell( task )

            # Compute times of interest.

            taskend = time.time()
            elapsed = taskend - taskstart

            # Send request for another task to the dispatcher. Include
            # some related information about the task just executed.

            task_socket.send_json( { 'mode' : 'ran',
                                     'worker' : workerID, 
                                     'tasknum' : tasknum,
                                     'task' : task,
                                     'status' : result[0],
                                     'stdout' : result[1],
                                     'stderr' : result[2],
                                     'tasktime' : elapsed,
                                     'taskstart' : taskstart,
                                     'taskend' : taskend } )

         else:

            running = False

      else:
         # Poll timed out??
         sys.stderr.write( "Worker:TimedOut:%s\n" % ( workerID ) )
         sys.stderr.flush()
         running = False

   task_socket.setsockopt( zmq.LINGER, 0 )
   task_socket.close()
   context.term()


def print_results( log, results, verbose ):
   """
   Procedure to print out results. The results argument is a dictionary
   whose contents varies by mode, but both have:

       'log' ..... file object to be written to.
       'mode' .... may be 'ran' or 'start
       'worker' .. the workerID (hostnm_#)

   ran:

       'tasknum' .... the task number completed
       'task' ....... the task string (line from input file)
       'status' ..... execution success or failure
       'stdout' ..... all stdout lines from task
       'stderr' ..... all stderr lines from task
       'tasktime' ... task elapsed time
       'taskstart' .. worker local start time
       'taskend' .... worker local end time

   start:

       'localtime' .. worker local start time
       'affinity' ... numactl bindings
       'cpt' ........ cores per thread string

   """
   worker = results['worker']
   if results['mode'] == 'start' :
      log.write( "Worker:Startup:%s:%.2f\n" \
                 % ( worker, results['localtime'] ) )
      if verbose:
         log.write( "Worker:Affinity:%s:%s\n" \
                    % ( worker, results['affinity'] ) )
         log.write( "Worker:CPT:%s:%s\n" % ( worker, results['cpt'] ) )
      log.flush()
   elif results['mode'] == 'ran' :
      if verbose :
         log.write( "Worker:Task:%d:%s:%s\n"
                    % ( results['tasknum'],
                        results['worker'],
                        results['task'] )
                    + "Worker:Timings:%d:%s:%.2f:%.2f:%.2f\n"
                    % ( results['tasknum'],
                        results['worker'],
                        results['taskstart'],
                        results['taskend'],
                        results['tasktime'] ) )
      log.write( "Worker:Stdout:%d:%s:%s\n" \
                 % ( results['tasknum'], worker, results['status'] ) )
      for l in results['stdout']:
         if len( l.strip() ) > 0:
            log.write( "  %s\n" % ( l.strip() ) )
      log.write( "Worker:Stderr:%d:%s\n" % (results['tasknum'], worker ) )
      for l in results['stderr']:
         if len( l.strip() ) > 0:
            log.write( "  %s\n" % ( l.strip() ) )
   log.flush()


def Usage():
   global jobtime
   sys.stderr.write( """
Usage:  wq.py opt1 [opt2 ...]

   As Dispatcher:

       wq.py -d[--dispatcher] cmd [-s[--start] task_num] \\
             [-t[--time] walltime] [-v[--verbose]] -a[--allworkers] n \\
             -i[--input] filenm -j[--jobid] jobid

   As Worker:

       wq.py -w[--workers] n -m[--mothersuperior] ms

   Help:

       wq.py -h[--help]

   General Options:

      -v,--verbose ..... Increase STDERR output details.
      -h,--help ........ Display this help message.

   Dispatcher Options:

      -d,--dispatcher cmd .. (Required) Run as the dispatcher for the
                             command cmd, where cmd is the task
                             program or script. cmd will be called
                             with a single file name as its only
                             argument.

      -a,--allworkers n .... (Required) Total workers ( workers per
                             node * nodes ).

      -i,--inputs filenm ... (Required) Name of file containing file
                             names to serve as inputs to cmd, one name
                             per line.

      -j,--jobid ........... (Option) PBS jobnumber. If provided, will
                             be added to the log file name: wq.log.jobid

      -s,--start task_num .. (Optional) Task number to start
                             with. Represents the line number in the
                             input list file. Default is 1.

      -t,--time walltime ... (Optional) Wallclock time to allow for
                             entire job.  May be expressed as one
                             of the following:

                                 ss, mm:ss, hh:mm:ss, or d:hh:mm:ss.

                             (note: Torque sets env variable PBS_WALLTIME)

   Worker Options:

      -w,--workers n .......... (Required) Run n workers per node.

      -m,--mothersuperior ms .. (Required) Host name of mother superior
                                node.

   The dispatcher must be started before any of the workers.\n
""" )
   sys.stderr.write( "   The default jobtime is hardwired to " \
                     + "%d secs - 1 day.\n\n" % ( jobtime ) )
   sys.stderr.write( "   Revision: $Id: wq.py 264 2016-01-25 16:22:13Z jalupo $\n" )


def time2secs( s ):
   """
   Convert time in a string to seconds.

   Args:

      s .. time string. May be 'ss', 'mm:ss', 'hh:mm:ss', 'd:hh:mm:ss'

   Returns:

      Number of seconds as an int.
   """
   t = s.split( ':' )
   nf = len( t )
   if nf == 1:
      # Seconds only!
      secs = int( t[0] )
   elif nf == 2:
      # Minutes & seconds!
      secs = int( t[1] ) + int( t[0] ) * 60
   elif nf == 3:
      # Hours, minutes & seconds!
      secs = int( t[2] ) + int( t[1] ) * 60 + int( t[0] ) * 60 * 60   
   elif nf == 4:
      # Days, hours, minutes, & seconds!
      secs = int( t[3] ) + int( t[2] ) * 60 + int( t[1] ) * 60 * 60
      secs += int( t[0] ) * 60 * 60 * 24

   return secs


# And here is the main line.

if __name__ == "__main__":

   global jobtime

   port = '54321'
   start = 1
   jobtime = 86400
   filenm = ''
   ms = ''
   verbose = False
   jobid = ''

   try:

      opts, args = getopt.getopt( sys.argv[1:], "hvd:w:s:i:t:m:a:j:",
                                  ['help', 'verbose', 'dispatcher=', 
                                   'workers=', 'start=', 'inputs=',
                                   'time=', 'mothersuperior=',
				   'allworkers=','jobid='] )

   except getopt.GetoptError, err:

      sys.stderr.write( str( err ) )
      Usage()
      sys.exit( 1 )

   for o, a in opts:

      if o in ( "-d", "--dispatcher" ) :

         mode = 'd'
         cmd = a

      elif o in ( "-w", "--workers" ) :

         mode = 'w'
         numw = int( a )

      elif o in ( "-a", "--allworkers" ) :

         allw = int( a )

      elif o in ( "-j", "--jobid" ) :

         jobid = a

      elif o in ( "-s", "--start" ) :

         start = int ( a )

      elif o in ( "-i", "--inputs" ) :

         filenm = a

      elif o in ( "-m", "--mothersuperior" ) :

         ms = a

      elif o in ( "-t", "--time" ) :

         jobtime = time2secs( a )

      elif o in ( "-h", "--help" ) :

         Usage()
         sys.exit( 0 )

      elif o in ( "-v", "--verbose" ) :

         verbose = True

      else:

	 sys.stderr.write( "ERROR: Unknown option: \"%s\"\n" % ( o ) \
                           + "Run with -h or --help for usage hints.\n" )
         sys.exit( 1 )

   if mode == 'w':

      if ms == '' :
         sys.stderr.write( "ERROR: Mother superior host name not specified.\n" \
                           + "Run with -h or --help for usage hints.\n" )
         sys.exit( 1 )

      if numw < 1 :
         sys.stderr.write( "ERROR: Number of workers must be positive! " \
                           + "Have: %d\n" % ( numw ) )
         sys.exit( 1 )

      # Determine the CPU affinity for the workers:

      bindings,cpt = physcpubindings( numw )

      # Get hostname of mother superior node.

      host = ipaddrs( ms )

      # Launch the desired number of worker threads.

      for wrk_num in range( numw ):

         Process( target = worker,
                  args = ( wrk_num, host, port, verbose,
                           bindings[wrk_num], cpt ) ).start()

   if mode == 'd':

      if jobtime < 1 :
         sys.stderr.write( "ERROR: --jobtime must be positive! Have: %d\n" \
                           % ( jobtime ) )
         sys.exit( 1 )

      if allw < 1 :
         sys.stderr.write( "ERROR: --allworkers must be positive! Have: %d\n" \
                           % ( allw ) )
         sys.exit( 1 )

      # Open the input list file.

      try:
         infile = open( filenm, 'r')
      except IOError:
         sys.stderr.write( "ERROR: Failed to open inputs file: \"%s\"\n" \
                           % ( filenm ) )
         sys.exit( 1 )

      lines = infile.readlines()
      infile.close()
      tasks = len( lines )

      # Fire up the dispatcher!

      if start > tasks:
         sys.stderr.write( "ERROR: Starting point (%d) exceeds input " \
                           + "lines (%d)!\n" % ( start, tasks ) )
         sys.exit( 1 )

      if tasks > 0:
         # Getting the process exit code is a bit tricky. We could run
         # the dispatcher as a subroutine, then there is no issue. That
         # would look like:
         #
         #     dispatcher( port, cmd, lines, allw, start, jobtime, verbose )
         #
         # For symmetry with the workers, let's run as a process then wait
         # for completion. This works because only 1 dispatcher is needed.
         dispatcher = Process( target = dispatcher,
                               args = ( port, cmd, lines, allw, start,
                                        jobtime, verbose, jobid ) )
         dispatcher.start()
         dispatcher.join()
         sys.exit( dispatcher.exitcode )
      else:
         sys.stderr.write( "ERROR: Inputs file appears empty: \"%s\"\n" \
                           % ( filenm ) )
         sys.exit( 1 )

# And we're out'a here!
